"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const graphql_1 = require("graphql");
const chalk_1 = __importDefault(require("chalk"));
const path_1 = require("path");
const fs_extra_1 = require("fs-extra");
const chokidar_1 = require("chokidar");
const glob = __importStar(require("glob"));
const graphql_config_1 = require("graphql-config");
const ast_1 = require("graphql-tool-utilities/ast");
const print_1 = require("./print");
const types_1 = require("./types");
exports.EnumFormat = types_1.EnumFormat;
class Builder extends events_1.EventEmitter {
    constructor(_a) {
        var { cwd } = _a, options = __rest(_a, ["cwd"]);
        super();
        this.documentMapByProject = new Map();
        this.options = options;
        this.config = graphql_config_1.getGraphQLConfig(cwd ? path_1.resolve(cwd) : undefined);
    }
    once(event, handler) {
        return super.once(event, handler);
    }
    on(event, handler) {
        return super.on(event, handler);
    }
    emit(event, ...args) {
        return super.emit(event, ...args);
    }
    run({ watch: watchGlobs = false } = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const globs = this.getGlobs();
            const schemaPaths = this.getSchemaPaths();
            const update = (filePath) => __awaiter(this, void 0, void 0, function* () {
                try {
                    yield this.updateDocumentForFile(filePath);
                }
                catch (error) {
                    this.emit('error', error);
                    return;
                }
                yield this.generateDocumentTypes();
            });
            if (watchGlobs) {
                const documentWatcher = chokidar_1.watch(globs);
                documentWatcher.on('ready', () => {
                    documentWatcher.on('add', update);
                    documentWatcher.on('change', update);
                    documentWatcher.on('unlink', (filePath) => __awaiter(this, void 0, void 0, function* () {
                        this.removeDocumentForFile(filePath);
                        yield this.generateDocumentTypes();
                    }));
                });
                const schemaWatcher = chokidar_1.watch(schemaPaths);
                schemaWatcher.on('ready', () => {
                    schemaWatcher.on('change', (schemaPath) => __awaiter(this, void 0, void 0, function* () {
                        try {
                            this.emit('start:schema');
                            yield this.updateSchemaAndGenerateTypes(schemaPath);
                            this.emit('end:schema');
                            yield this.generateDocumentTypes();
                        }
                        catch (error) {
                            // intentional noop
                        }
                    }));
                });
            }
            try {
                this.emit('start:schema');
                yield Promise.all(schemaPaths.map(this.updateSchemaAndGenerateTypes.bind(this)));
                this.emit('end:schema');
            }
            catch (error) {
                this.emit('error', error);
                return;
            }
            try {
                yield Promise.all(globs
                    .map((pattern) => glob.sync(pattern))
                    .reduce((patterns, globbed) => patterns.concat(globbed), [])
                    .map(this.updateDocumentForFile.bind(this)));
            }
            catch (error) {
                this.emit('error', error);
                return;
            }
            yield this.generateDocumentTypes();
        });
    }
    generateSchemaTypes(schemaPath, schema) {
        return __awaiter(this, void 0, void 0, function* () {
            const schemaTypesPath = getSchemaTypesPath(getProjectForSchemaPath(this.config, schemaPath), this.options);
            const definition = print_1.printSchema(schema, this.options);
            yield fs_extra_1.mkdirp(path_1.dirname(schemaTypesPath));
            yield fs_extra_1.writeFile(schemaTypesPath, definition);
            this.emit('build:schema', {
                schemaPath,
                schemaTypesPath,
            });
        });
    }
    generateDocumentTypes() {
        return __awaiter(this, void 0, void 0, function* () {
            this.emit('start:docs');
            getDuplicateOperations(this.documentMapByProject).forEach(({ projectName, duplicates }) => {
                if (duplicates.length) {
                    duplicates.forEach(({ operationName, filePaths }) => {
                        const message = `GraphQL operations must have a unique name. The operation ${chalk_1.default.bold(operationName)} is declared in:\n ${filePaths.sort().join('\n ')}${projectName ? ` (${chalk_1.default.bold(projectName)})` : ''}`;
                        this.emit('error', new Error(message));
                    });
                }
            });
            yield Promise.all(Array.from(this.documentMapByProject.entries()).map(([projectName, documents]) => {
                return this.generateDocumentTypesForProject(this.config.getProjectConfig(projectName), documents);
            }));
            this.emit('end:docs');
        });
    }
    generateDocumentTypesForProject(project, documents) {
        return __awaiter(this, void 0, void 0, function* () {
            let ast;
            try {
                const schema = project.getSchema();
                ast = ast_1.compile(schema, graphql_1.concatAST(Array.from(documents.values())));
            }
            catch (error) {
                this.emit('error', error);
                return;
            }
            const fileMap = groupOperationsAndFragmentsByFile(ast);
            try {
                const buildResults = yield Promise.all(Array.from(fileMap.values()).map((file) => {
                    return this.writeDocumentFile(file, ast, project);
                }));
                for (const buildResult of buildResults) {
                    this.emit('build:docs', buildResult);
                }
            }
            catch (error) {
                // intentional noop
            }
        });
    }
    writeDocumentFile(file, ast, project) {
        return __awaiter(this, void 0, void 0, function* () {
            const definitionPath = `${file.path}.d.ts`;
            const definition = this.getDocumentDefinition(file, ast, project);
            yield fs_extra_1.writeFile(definitionPath, definition);
            return {
                documentPath: file.path,
                definitionPath,
                operation: file.operation,
                fragments: file.fragments,
            };
        });
    }
    getDocumentDefinition(file, ast, project) {
        try {
            return print_1.printDocument(file, ast, Object.assign({}, this.options, { schemaTypesPath: getSchemaTypesPath(project, this.options) }));
        }
        catch ({ message }) {
            const error = new Error(`Error in ${file.path}: ${message[0].toLowerCase()}${message.slice(1)}`);
            this.emit('error', error);
            throw error;
        }
    }
    updateSchemaAndGenerateTypes(schemaPath) {
        return __awaiter(this, void 0, void 0, function* () {
            const schema = yield this.updateSchema(schemaPath);
            yield this.generateSchemaTypes(schemaPath, schema);
        });
    }
    updateSchema(schemaPath) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const schemaJSON = yield fs_extra_1.readJSON(schemaPath);
                return graphql_1.buildClientSchema(schemaJSON.data);
            }
            catch (error) {
                const parseError = new Error(`Error parsing '${schemaPath}':\n\n${error.message.replace(/Syntax Error GraphQL \(.*?\) /, '')}`);
                throw parseError;
            }
        });
    }
    updateDocumentForFile(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            const project = this.config.getConfigForFile(filePath);
            if (!project) {
                throw new Error(`No project found for file: ${filePath}`);
            }
            const projectName = project.projectName || '';
            let documents = this.documentMapByProject.get(projectName);
            if (!documents) {
                documents = new Map();
                this.documentMapByProject.set(projectName, documents);
            }
            const contents = yield fs_extra_1.readFile(filePath, 'utf8');
            if (contents.trim().length === 0) {
                return undefined;
            }
            const document = graphql_1.parse(new graphql_1.Source(contents, filePath));
            documents.set(filePath, document);
            return document;
        });
    }
    getProjects() {
        const projects = this.config.getProjects();
        if (projects) {
            return Object.values(projects);
        }
        return [this.config.getProjectConfig()];
    }
    getGlobs() {
        return this.getProjects().reduce((globs, project) => {
            return globs.concat(project.includes.map(getFilePath.bind(this, this.config.configDir)));
        }, []);
    }
    getSchemaPaths() {
        return this.getProjects().reduce((schemas, x) => {
            if (x.schemaPath) {
                schemas.push(getFilePath(this.config.configDir, x.schemaPath));
            }
            return schemas;
        }, []);
    }
    removeDocumentForFile(filePath) {
        const project = this.config.getConfigForFile(filePath);
        if (project) {
            const projectName = project.projectName || '';
            const documents = this.documentMapByProject.get(projectName);
            if (documents) {
                documents.delete(filePath);
            }
        }
    }
}
exports.Builder = Builder;
function getFilePath(cwd, filePath) {
    return path_1.isAbsolute(filePath) ? filePath : path_1.resolve(cwd, filePath);
}
function getProjectForSchemaPath(config, schemaPath) {
    const project = Object.values(config.getProjects() || {})
        .filter((project) => project.schemaPath === schemaPath)
        .shift() || config.getProjectConfig();
    if (!project || project.schemaPath !== schemaPath) {
        throw new Error(`No project defined in .graphqlconfig for schema ${schemaPath}`);
    }
    return project;
}
function getSchemaTypesPath(project, options) {
    if (typeof project.extensions.schemaTypesPath === 'string') {
        return getFilePath(project.configDir, project.extensions.schemaTypesPath);
    }
    return getFilePath(project.configDir, path_1.resolve(options.schemaTypesPath, `${project.projectName ? `${project.projectName}-` : ''}types.ts`));
}
function groupOperationsAndFragmentsByFile({ operations, fragments }) {
    return Object.values(operations)
        .concat(Object.values(fragments))
        .reduce((map, item) => {
        let file = map.get(item.filePath);
        if (!file) {
            file = {
                path: item.filePath,
                operation: undefined,
                fragments: [],
            };
            map.set(item.filePath, file);
        }
        if (ast_1.isOperation(item)) {
            file.operation = item;
        }
        else {
            file.fragments.push(item);
        }
        return map;
    }, new Map());
}
function getDuplicateOperations(documentsMapByProject) {
    return Array.from(documentsMapByProject.entries()).map(([projectName, documents]) => {
        return {
            projectName,
            duplicates: getDuplicateProjectOperations(documents),
        };
    });
}
function getDuplicateProjectOperations(documents) {
    const operations = new Map();
    Array.from(documents.entries()).forEach(([filePath, document]) => {
        document.definitions.filter(isOperationDefinition).forEach((definition) => {
            const { name } = definition;
            if (name && name.value) {
                const map = operations.get(name.value);
                if (map) {
                    map.add(filePath);
                }
                else {
                    operations.set(name.value, new Set([filePath]));
                }
            }
        });
    });
    return Array.from(operations.entries())
        .filter(([, filePaths]) => filePaths.size > 1)
        .map(([operationName, filePath]) => {
        return { operationName, filePaths: Array.from(filePath) };
    });
}
function isOperationDefinition(definition) {
    return definition.kind === 'OperationDefinition';
}
